# Configuation for the test environment

# This file has the global configuration 
# so we do not have to repeat those in every function 

import pandas as pd
import os, sys
import datetime

# This is for production environment
# The tests will define these paths separately
print(" Test Config imported")

dirname = os.path.dirname(os.path.dirname(__file__))
###################################################
# This is the main change from the production config 
DATA_ROOT = os.path.join(dirname, "data/measurements")
######################################################
print(DATA_ROOT)
METADATA_PATH = os.path.join(dirname, "data/properties/")
METER_CHANNEL_FILE = os.path.join(METADATA_PATH, "NetCDF Meter File Generation Matrix Copy Poker Flats.xlsx")
DATA_YEARS_FILE = os.path.join(METADATA_PATH, "data_years.txt")
METER_NAMES_FILE = os.path.join(METADATA_PATH, "meter_names.txt")

# Get meter names from meter names file 
meter_names_df = pd.read_csv(METER_NAMES_FILE)
print(meter_names_df)
METER_NAMES = meter_names_df['meter_name'].tolist()

# Read in files containing channel names for WattsOn
meter_channel_metadata_WattsOnMetadata = pd.read_excel(METER_CHANNEL_FILE, sheet_name="WattsOnMk2")
meter_channel_metadata_PQube = pd.read_excel(METER_CHANNEL_FILE, sheet_name="PQube3PF")

# Extract channel names

###########################################################################
# TODO: Create a metadata file with channel names for every meter 
# ----------------------------------------------------------------
# this is hardcoded to read upto line 48, 
# as the file contains other lines at the end. This should be changed to a 
# file containing channels per meter, as different meters can have different
# channels and that file should be a source of truth across applications, 
# the new database schema is created when this file changes
############################################################################

# # Additional column name time is added to store the timestamp of measurement
# channel_names = ['time'] + list(meter_details['Channels'][:48])
# # Extract name of meters 
# # TODO: Change this part to remmove the hardcoding
# #  when the metadata files are sorted
# # ---------------------------------------------------------
# meter_names = list(meter_details.columns.values)[-4:]

# Create a dictionary to store the channels per meter. So keys are the meter 
# names and the values are a list of channel per meter
METER_CHANNEL_DICT = {}
# Loop across the meter_names list to add channels for each meter
for index, row in meter_names_df.iterrows():
    if row['meter_type'] == 'WattsOnMk2':
        channel_names = meter_channel_metadata_WattsOnMetadata['Filename'][0:48]
        METER_CHANNEL_DICT[row['meter_name']] = ['time'] + list(channel_names) 
    elif row['meter_type'] == 'PQube':
        channel_names = meter_channel_metadata_PQube['Filename'][0:46]
        METER_CHANNEL_DICT[row['meter_name']] = ['time'] + list(channel_names)
        
# Get years from the years file
years_df = pd.read_csv(DATA_YEARS_FILE)

DATA_YEARS = years_df['years'].values.tolist()

# Data start and end date
DATA_START_DATE = datetime.datetime(2017, 11, 1)
DATA_END_DATE = datetime.datetime(2019, 4, 30)

# Database name 
DB_NAME = 'demand_acep'

# Downsampling duration
# sample_time allows the user determine what time interval the data should be resampled at
# For 1 minute - 1T, 1 hour - 1H, 1 month - 1M, 1 Day - 1D
SAMPLE_TIME = '1T'

# Data imputation related settings
# Data Imputation by interpolation
# interp_method and interp_order allows the user specify the method of interpolation and the order
INTERP_METHOD = 'spline'
INTERP_ORDER = 2


from .version import __version__  # noqa
from .demand_acep import * # noqa
from .create_db_schema import *
# from .config import *
from .extract_data_to_csv import *
from .timescale_parallel_copy import *


"""
This is the place for all the unit tests for the package demand_acep.
We can refactor this if this becomes too large.

"""
# %% Imports

import os
import sys
import numpy as np
import pandas as pd
import pdb
import pytest
import copy
import importlib
import unittest

from itertools import groupby
from operator import itemgetter

# To import files from the parent directory
sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from demand_acep import extract_data
from demand_acep import extract_ppty
from demand_acep import data_resample
from demand_acep import data_impute
from demand_acep import compute_interpolation
from demand_acep import build_interpolation
from demand_acep import long_missing_data_prep
from extract_data_to_csv import extract_csv_for_date
# %% Paths
path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
# dirpath = os.path.join(path, 'data/measurements/2018/07/01')
# filename = 'PokerFlatResearchRange-PokerFlat-PkFltM1AntEaDel@2018-07-02T081007Z@PT23H@PT146F.nc'
dirpath = os.path.join(path, 'data/measurements/2019/01/03')
filename = 'PokerFlatResearchRange-PokerFlat-PkFltM3SciEaDel@2019-01-03T093004Z@P1D@PT179F.nc'

# Test config 
import test_config as config

def test_extract_data():
    test_df = extract_data(dirpath, filename)
    column_name = test_df.columns.tolist()[0]

    assert (test_df.index.dtype == 'datetime64[ns]'), "The first output from this function should be a timedelta object"

    assert (test_df[column_name].dtype == 'float64'), "The second output from this function should be numpy array"

    return


def test_extract_ppty():
    meter_name = ['PkFltM1Ant', 'PkFltM2Tel', 'PkFltM3Sci', 'PQube3']
    [test_meter, test_channel] = extract_ppty(filename, meter_name)

    assert (any(val == test_meter for val in meter_name)), "Returned meter name does not exist"

    assert (test_channel in filename), "Returned measurement channel does not exist"

    return


def test_data_resample():
    test_df = extract_data(dirpath, filename)
    test_resampled = data_resample(test_df, sample_time='1T')
    diff_test = np.diff(test_resampled.index)
    time_1T_ns = np.timedelta64(60000000000,'ns') # sample_time 1T in nanoseconds
    assert (np.all(np.equal(diff_test, time_1T_ns))), "Data not properly downsamples"

    return


def test_data_impute():
    test_df = extract_data(dirpath, filename)
    test_df = data_impute(test_df)
    dict_assert = []
    if isinstance(test_df, dict):
        for meter in test_df:
            dict_assert.append(test_df[meter].notnull().values.all())
        assert (all(val for val in dict_assert)), "Data imputations in dictionary not functioning properly as data " \
                                                  "still contains NaN"
    else:
        assert (test_df.notnull().values.all()), "Data imputations in Dataframes not functioning properly as data " \
                                                 "still contains NaN"

    return

def test_extract_csv_for_date_badIn():
    
    # Test that bad input throws the right kind of exceptions 
    
    # DATA_ROOT attribute does not exist
    config_1 = importlib.reload(config)
    vars(config_1).pop('DATA_ROOT', None)
    # Correct date 
    data_date_1 = "07/01/2018"
    
    #print(config_1.DATA_ROOT)
    
    with pytest.raises(AttributeError):
        extract_csv_for_date(config_1, data_date_1)
    
    config_2 = importlib.reload(config)
    # Bad data_root
    config_2.DATA_ROOT = "/not_a_directory"
    # Correct date 
    data_date_2 = "07/01/2018"
    
    #print(config.DATA_ROOT)
    
    with pytest.raises(NotADirectoryError):
        extract_csv_for_date(config_2, data_date_2)
    
    config_3 = importlib.reload(config)
    data_date_3 = "07/01/2018"
    # No meter_channel_dict
    vars(config_3).pop('METER_CHANNEL_DICT', None)
    
    with pytest.raises(AttributeError):
        extract_csv_for_date(config_3, data_date_3)
    
    config_4 = importlib.reload(config)
    data_date_4 = "07/01/2018"
    
    # No SAMPLE_TIME
    vars(config_4).pop('SAMPLE_TIME', None)
    
    with pytest.raises(AttributeError):
        extract_csv_for_date(config_4, data_date_4)
        
    
    config_5 = importlib.reload(config)
    data_date_5 = "10/01/2017"
    
    with pytest.raises(ValueError):
        extract_csv_for_date(config_5, data_date_5)
    
    return 


def test_extract_csv_for_date():
    # Test date 
    test_data_date = '2019/01/03'
    # Re-import the config to make changes for this test
    test_config1 = importlib.reload(config)
    # Get the new DATA_ROOT to the location of the test data
    test_config1.DATA_ROOT = os.path.join(test_config1.DATA_ROOT, 'part_data')
    # Run the test function with the inputs
    test_csv_names = extract_csv_for_date(test_config1, test_data_date)
    print(test_csv_names)
    # Test the output filenames
    # Get correct names
    correct_csv_path = os.path.join(test_config1.DATA_ROOT, test_data_date, "correct_csvs")
    print(correct_csv_path)
    correct_names = [f for f in os.listdir(correct_csv_path) if os.path.isfile(os.path.join(correct_csv_path, f))]
    print(correct_names)
    # test_names_result = all(elem in correct_names for elem in test_csv_names)
    assert(checkEqual(correct_names, list(test_csv_names.values())), "The output csv names do not match the expected name format - meter_name@date@freq.csv")

    # Test the output values
    # Read in the correct csvs 
    correct_dfs = {}
    for dirpath, dirnames, correct_csvfiles in os.walk(correct_csv_path, topdown=True):
        for csvfilename in correct_csvfiles:
            csvfile_abspath = os.path.join(correct_csv_path, csvfilename)
            correct_dfs[csvfilename] = pd.read_csv(csvfile_abspath, header=None)
    
    test_csv_path = os.path.join(test_config1.DATA_ROOT, test_data_date)
    test_dfs = {}
    for dirpath, dirnames, test_files in os.walk(test_csv_path, topdown=True):
        for filename in test_files:
            if filename.lower().endswith('.csv'):
                csvfile_abspath = os.path.join(test_csv_path, filename)
                test_dfs[filename] = pd.read_csv(csvfile_abspath, header=None)
                # Remove the NA columns - these are introduced since we have only two channels
                test_dfs[filename].dropna(axis=1, inplace=True)

    # Check if the dataframes in correct_dfs and test_dfs are equal 
    for csvname in correct_dfs:
        correct_df = correct_dfs[csvname]
        test_df = test_dfs[csvname]
        pd.testing.assert_frame_equal(test_df, correct_df)

    return 


def test_build_interpolation():
    df = extract_data(dirpath, filename)
    test_resampled = data_resample(df, sample_time='1T')
    test_df = test_resampled.copy()
    # gets the index location in integers where the NaNs are located
    get_nan_idx = np.where(test_df.isna())[0]
    idx_grp_nan = []
    # creates a list of consecutive index locations to determine the range of interpolation
    for k, g in groupby(enumerate(get_nan_idx), lambda ix: ix[0] - ix[1]):
        idx_grp_nan.append(list(map(itemgetter(1), g)))
    # performs interpolation for each consecutive NaN index location
    for idx, val in enumerate(idx_grp_nan):
        n_grp = len(val)
        # for each range of consecutive NaN locations, use data points of length equal to the number that NaNs in that
        # range before and after the NaN datapoints
        prev_idx = val[0] - n_grp
        next_idx = val[-1] + n_grp
        # This if-else clause handles edge cases
        # If - When the number of consecutive NaN points is larger than the number of available data points before it in
        # the dataframe.
        # Elif - When the number of consecutive NaN points is larger than the number of available data points after it
        # in the dataframe.
        if prev_idx < 0:
            prev_vals = test_df.iloc[0:val[0]]
            next_vals = test_df.iloc[val[0] + 1: next_idx + 1]
        elif next_idx > len(test_df):
            prev_vals = test_df.iloc[prev_idx:val[0]]
            next_vals = test_df.iloc[val[0] + 1: len(test_df)]
        else:
            prev_vals = test_df.iloc[prev_idx:val[0]]
            next_vals = test_df.iloc[val[0] + 1: next_idx + 1]
        y_values = prev_vals.append(next_vals)
    y_interp = build_interpolation(y_values, n_grp)
    # pdb.set_trace()
    # check that y_interp is has no NaN values
    assert (~np.isnan(y_interp).any()), "Interpolation function not fully replacing NaN with interpolation values"
    # check that y_interp is within reasonable range from y_values
    max_val = np.max(y_values) + np.std(y_values)
    min_val = np.min(y_values) - np.std(y_values)
    assert (y_interp.all() < max_val).all() or (y_interp.all() > min_val).all(), "Error in determining a function fit " \
                                                                                 "through interpolation"

    assert (y_interp.dtype == 'float64'), "The output from this function should be numpy array"
    return


def test_compute_interpolation():
    test_df = extract_data(dirpath, filename)
    if isinstance(test_df, dict):
        for meter in test_df:
            assert (isinstance(test_df[meter], pd.DataFrame)), "Object passed in is not a Dataframe"
            if test_df[meter].isnull().values.any():
                test_df[meter] = test_df[meter].apply(compute_interpolation)
                assert (test_df[meter].notnull().values.all()), "Data imputations in Series not functioning properly " \
                                                                "as data still contains NaN"
    else:
        assert (isinstance(test_df, pd.DataFrame)), "Object passed in is not a Dataframe"
        if test_df.isnull().values.any():
            test_df = test_df.apply(compute_interpolation)
            assert (test_df.notnull().values.all()), "Data imputations in Series not functioning properly as data " \
                                                     "still contains NaN"

    return


def test_long_missing_data_prep():
    dirpath_data = os.path.join(path, 'data/measurements/test_data')
    filename_data = 'PQube3_comb.csv'

    meter_mod_df = long_missing_data_prep(dirpath_data, filename_data)

    assert (filename_data[-4:] == '.csv'), "File passed in is not in a .csv format"
    assert (isinstance(meter_mod_df, pd.DataFrame)), "Object returned in is not a Dataframe"
    meter_df = pd.read_csv(os.path.join(dirpath_data, filename_data))
    assert (len(meter_mod_df) >= len(meter_df)), "Returned DataFrame should be greater than or equal to the data " \
                                                 "passed in"
    diff_test = np.diff(meter_mod_df.index)
    time_1T_ns = np.timedelta64(60000000000, 'ns')  # sample_time 1T in nanoseconds
    assert (np.all(np.equal(diff_test, time_1T_ns))), "DateTimeIndex intervals not properly computed in function"

    return

# Taken from https://stackoverflow.com/a/12813909/1328232
# To check the equality of two lists
def checkEqual(L1, L2):
    return len(L1) == len(L2) and sorted(L1) == sorted(L2)



from os.path import join as pjoin

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
_version_major = 0
_version_minor = 0
_version_micro = '1'  # use '' for first of series, number for 1 and above
_version_extra = ''
# _version_extra = ''  # Uncomment this for full releases

# Construct full version string from these.
_ver = [_version_major, _version_minor]
if _version_micro:
    _ver.append(_version_micro)
if _version_extra:
    _ver.append(_version_extra)

__version__ = '.'.join(map(str, _ver))

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Description should be a one-liner:
description = "demand_acep : A package for demand charge reduction for ACEP"
# Long description will go up on the pypi page
long_description = "Python package to implement data-pipeline to process high-resolution power meter data. For documentation refer to: https://demand-acep.readthedocs.io/en/latest/"

NAME = "demand_acep"
MAINTAINER = "Chintan Pathak, Yohan Min, Atinuke Ademola-Idowu"
MAINTAINER_EMAIL = "chintan.pathak@gmail.com"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "http://github.com/demand-consults/demand_acep"
DOWNLOAD_URL = ""
LICENSE = "MIT"
AUTHOR = "Chintan Pathak, Yohan Min, Atinuke Ademola-Idowu"
AUTHOR_EMAIL = "cp84@uw.edu, min25@uw.edu, aidowu@uw.edu"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGE_DATA = {'demand_acep': [pjoin('data', '*')]}
REQUIRES = ["pandas", "numpy", "sklearn", "pytest"]



""" The functions in this file convert the netcdf files to a dataframe and then 
write them to disk using multiprocessing. So, if we have a multi-core system, the 
extraction can happen in parallel. """

import os
import pandas as pd
import numpy as np
import pdb
import datetime
import io
import multiprocessing as mp

from demand_acep import extract_data
from demand_acep import extract_ppty
from demand_acep import data_resample
from demand_acep import data_impute
from dateutil.parser import parse


def extract_csv_for_date(config, data_date):
    """ 
    This function writes the data for the data_date specified to a csv on disk.
    
    Parameters
    ----------
    config : 
        `config` contains the configuration for paths etc. needed by files. As
        an argument we can change the config files for different production 
        vs test configs 
    data_date : string
        `data_date` string will be used to extract the year and the path to the
        data. It should be in the 'mm/dd/yyyy' format. 

    Returns
    -------
    list
        Names of the csv's written to disk.
    """    
    
    ### TODO: test config separately 
    
    # print(config.DATA_ROOT)
    # print(data_date)
    
    # Raise an exception if attribute DATA_ROOT does not exist
    if not 'DATA_ROOT' in vars(config):
        raise AttributeError("Attribute DATA_ROOT does not exist")
        
    # Raise an exception if DATA_ROOT does not exist
    if not os.path.exists(config.DATA_ROOT):
        raise NotADirectoryError("The path " + config.DATA_ROOT + " not found")
        
    # Raise an exception if attribute METER_CHANNEL_DICT does not exist
    if not 'METER_CHANNEL_DICT' in vars(config):
        raise AttributeError("Attribute METER_CHANNEL_DICT does not exist")
        
    # Raise an exception if attribute METER_CHANNEL_DICT does not exist
    if not 'SAMPLE_TIME' in vars(config):
        raise AttributeError("Attribute METER_CHANNEL_DICT does not exist")
    
    data_date_dt = parse(data_date)
    
    if data_date_dt > config.DATA_END_DATE:
        raise ValueError("data_date entered is greater than the DATA_END_DATE: " + 
                        str(config.DATA_END_DATE))
                        
    if data_date_dt < config.DATA_START_DATE:
        raise ValueError("data_date entered is less than the DATA_START_DATE: " + 
                        str(config.DATA_START_DATE))
                        
    # Get the year, month and and day from date entered
    data_year = data_date_dt.year
    data_month = data_date_dt.month
    data_day = data_date_dt.day
    
    # Get the corresponding path in the directory to look for the data for the day
    data_path = os.path.join(config.DATA_ROOT, str(data_year), "{:02}".format(data_month), "{:02}".format(data_day))
    # print(data_path)
    # Find the count of meters
    meter_count = len(config.METER_CHANNEL_DICT)

    # Dictionary to store the names of the resulting csv files
    meter_csv_names = {}
    
    # Get the down-sampling time
    sample_time = config.SAMPLE_TIME
    
    # Create a dictionary with keys are meter names and values as dataframes 
    # containing the data for the day
    meter_collection = {}
    
    # for meter_name in config.METER_CHANNEL_DICT:
    #     # Create an empty dataframe, the columns will be created later
    #     meter_collection[meter_name] = pd.DataFrame()

    #print(meter_collection)
    if os.path.exists(data_path):
        # Walk through all the files in the directory for the day's data
        for dirpath, dirnames, files in os.walk(data_path, topdown=True):
            # `files` contains the names of all the files at the location
            if len(files) == 0:
                print("No files found for day: " + data_path)
                continue
            for filename in files:
                # Get the netcdf files, these are files with `.nc` extension
                if filename.lower().endswith('.nc'):
                    # For the particular file, find out the corresponding meter and channel 
                    [meter, channel] = extract_ppty(filename, config.METER_CHANNEL_DICT.keys())
                    # Create an entry in the `meter_collection` dict if it does not exist yet
                    if meter not in meter_collection:
                        meter_collection[meter] = pd.DataFrame()
                    # Form the resulting csv name from the meter name if it doesnt exist yet
                    # They are of the type - meter_name@Timestamp@Duration@Frequency
                    # For e.g.: PQube3@2017-11-01T080002Z@PT23H@PT227F.cs
                    #print(meter, channel)
                    if meter not in meter_csv_names:
                        meter_csv_names[meter] = '@'.join([meter, '@'.join(filename.split('@')[1:4])])[:-3] + '.csv'
                    #print(meter_csv_names)
                    # Get the full path of the csv
                    csv_name = os.path.join(data_path, meter_csv_names[meter])
                    # Only extract if not already extracted to csv
                    if (not os.path.isfile(csv_name)):
                        # Get the dataframe containing time and channel values
                        channel_df = extract_data(dirpath, filename)
                        # Give the dataframe column a name
                        channel_df.columns = [channel]
                        # Down-sample the data to the sampling time intended
                        channel_resampled = data_resample(channel_df, sample_time)
                        # If our meter dataframe is empty so far, i.e. if this is the 
                        # first channel being entered, then create a copy of the 
                        # resampled dataframe
                        if meter_collection[meter].empty:
                            meter_collection[meter] = channel_resampled.copy()
                        ####################### 
                        # This `else` clause handles two cases:
                        # 1. If the dataframe is not empty, then add other columns to
                        #    the dataframe. (the else case)
                        # 2. Some days have data downloaded more than once, this means 
                        #    that channels can occur more than once. (like 05/21/2018)
                        #######################
                        else:
                            # If the channel already exists in the dataframe
                            # then either the other file has updated data or 
                            # subsequent data. 
                            if channel in meter_collection[meter].columns:
                                # Get index from total dataframe 
                                idx_1 = meter_collection[meter].index
                                # Get index from file dataframe
                                idx_2 = channel_resampled.index
                                # Compare the two, if the index is contained within,
                                # then **update** the channel's value for file's indices. 
                                if np.all(np.isin(idx_2, idx_1)):
                                    meter_collection[meter][channel].loc[idx_2] = channel_resampled.values.tolist()
                                # If the index is not contained, append the file df to
                                # the total dataframe
                                else:
                                    meter_collection[meter] = meter_collection[meter].append(channel_resampled, sort=False)
                                    meter_collection[meter].sort_index(inplace=True)
                                    #######################
                                    # This data is resampled a second time to handle two cases:
                                    # 1. When appending a resampled dataframe to an already resampled dataframe, the last
                                    #    index of the original dataframe and the first index of the new dataframe can have
                                    #    the same time. Resampling the appended dataframe will eliminate the repetitions.
                                    # 2. If the new dataframe to be appended starts at a much later time, resampling the
                                    #    appended dataframe will create rows of missing data (NaN) at the times with no
                                    #    measurement values. This makes it easier to detect missing measurement values and
                                    #    perform data imputation at a later phase.
                                    #######################
                                    meter_collection[meter] = data_resample(meter_collection[meter], sample_time)
                            # If the channel does not already exist, then add the
                            # file dataframe to the total df. 
                            else:
                                meter_collection[meter] = meter_collection[meter].join(channel_resampled, how='outer')
    else:
        print("Path not found: " + data_path)
                
    # Set the data imputation parameters
    interp_method = config.INTERP_METHOD
    interp_order = config.INTERP_ORDER
    
    # Perform data imputation wherrever needed
    # print(meter_collection)
    meter_collection = data_impute(meter_collection)
    
    # Write the total dataframes to csv file
    for meter in meter_collection:
        # Reorganize the order of columns to match the database tables 
        meter_channels = config.METER_CHANNEL_DICT[meter]
        # meter_collection[meter].reset_index(inplace=True)
        meter_collection[meter] = meter_collection[meter].reindex(columns=meter_channels[1:])
        csv_name = os.path.join(data_path, meter_csv_names[meter])
        # print(csv_name)
        # Only write csv if it does not exist yet
        if(not os.path.isfile(csv_name)):
            meter_collection[meter].to_csv(csv_name, header=False)

    return meter_csv_names

def printResult(result):
    # print(result)
    return
    


""" 
This file has the global configuration so we do not have to repeat those in every function. 

We can create separate configuration files for production and test system, but they should atleast define the following:

Attributes
----------
DATA_ROOT: string 
    This is the absolute path of data. 

METADATA_PATH: string
    This is the absolute path of the metadata for our data. The metadata contains information about channels per meter, names and years of meters etc. 
    Sample metadata files are found in `https://github.com/demand-consults/demand_acep/tree/master/demand_acep/data/properties`.
    
METER_CHANNEL_FILE: string
    This is the absolute path of the files containing the names of channels for each meter. Each meter can have different channels it reads. 
    An example `METER_CHANNEL_FILE` is here: https://github.com/demand-consults/demand_acep/blob/master/demand_acep/data/properties/NetCDF%20Meter%20File%20Generation%20Matrix%20Copy%20Poker%20Flats.xlsx
    If the format of this file changes, then the parsing logic below will have to change accordingly.

DATA_YEARS_FILE: string 
    This is the absolute path of the data years file. This file lists all the years that we have the data for. The code then creates a new table for each meter for each year. 
    A sample data years file is here: https://github.com/demand-consults/demand_acep/blob/master/demand_acep/data/properties/data_years.txt

METER_NAMES_FILE: string 
    This is the absolute path of the meter names file. This file lists the names of the all the meters and their type. 
    A sample `METER_NAMES_FILE` is here: https://github.com/demand-consults/demand_acep/blob/master/demand_acep/data/properties/meter_names.txt

METER_NAMES: list
    This list contains the names of the meters. 

METER_CHANNEL_DICT: dictionary
    This dictionary should contain the available meter names as keys and the channels for that meter as values for the corresponding. 
    The generation of this dictionary is related to the structure of the `METER_CHANNEL_FILE`.

DATA_YEARS: list
    This list contains the years of the data. 

DATA_START_DATE: datetime.datetime
    The start date for the data to process. 

DATA_END_DATE: datetime.datetime
    The end date for the date to process. 

DB_NAME: string
    The name of the database to create the schema in. This database will store the processed data, with one table per meter per year. 

tsdb_pc_path: string
    The absolute path of the timescaledb-parallel-copy Go executable.

DB_USER: string 
    The username to connect to the TimescaleDB with

DB_PASSWORD: string 
    The password for the TimescaleDB database. 

SAMPLE_TIME: string 
    The argument needed to downsample the data, 1T = 1 min etc. Please refer to the pandas `resample` documentation here: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.resample.html


 """

import pandas as pd
import os, sys
import datetime

# This is for production environment
# The tests will define these paths separately
print("Config imported")
DATA_ROOT = "/gscratch/stf/demand_acep/Data"

dirname = os.path.dirname(__file__)
METADATA_PATH = os.path.join(dirname, "data/properties/")
METER_CHANNEL_FILE = os.path.join(METADATA_PATH, "NetCDF Meter File Generation Matrix Copy Poker Flats.xlsx")
DATA_YEARS_FILE = os.path.join(METADATA_PATH, "data_years.txt")
METER_NAMES_FILE = os.path.join(METADATA_PATH, "meter_names.txt")

# Get meter names from meter names file 
meter_names_df = pd.read_csv(METER_NAMES_FILE)
print(meter_names_df)
METER_NAMES = meter_names_df['meter_name'].tolist()

# Read in files containing channel names for WattsOn
meter_channel_metadata_WattsOnMetadata = pd.read_excel(METER_CHANNEL_FILE, sheet_name="WattsOnMk2")
meter_channel_metadata_PQube = pd.read_excel(METER_CHANNEL_FILE, sheet_name="PQube3PF")

# Extract channel names

###########################################################################
# TODO: Create a metadata file with channel names for every meter 
# ----------------------------------------------------------------
# this is hardcoded to read upto line 48, 
# as the file contains other lines at the end. This should be changed to a 
# file containing channels per meter, as different meters can have different
# channels and that file should be a source of truth across applications, 
# the new database schema is created when this file changes
############################################################################

# # Additional column name time is added to store the timestamp of measurement
# channel_names = ['time'] + list(meter_details['Channels'][:48])
# # Extract name of meters 
# # TODO: Change this part to remmove the hardcoding
# #  when the metadata files are sorted
# # ---------------------------------------------------------
# meter_names = list(meter_details.columns.values)[-4:]

# Create a dictionary to store the channels per meter. So keys are the meter 
# names and the values are a list of channel per meter
METER_CHANNEL_DICT = {}
# Loop across the meter_names list to add channels for each meter
for index, row in meter_names_df.iterrows():
    if row['meter_type'] == 'WattsOnMk2':
        channel_names = meter_channel_metadata_WattsOnMetadata['Filename'][0:48]
        METER_CHANNEL_DICT[row['meter_name']] = ['time'] + list(channel_names) 
    elif row['meter_type'] == 'PQube':
        channel_names = meter_channel_metadata_PQube['Filename'][0:46]
        METER_CHANNEL_DICT[row['meter_name']] = ['time'] + list(channel_names)
        
# Get years from the years file
years_df = pd.read_csv(DATA_YEARS_FILE)

DATA_YEARS = years_df['years'].values.tolist()


# Data start and end date
DATA_START_DATE = datetime.datetime(2017, 11, 1)
DATA_END_DATE = datetime.datetime(2019, 4, 30)

#######################################################
############ Database related configuration ###########
#######################################################


# Datebase IP address
DB_ADDRESS = "localhost"
# Database port 
DB_PORT = 5432
################################################################
############### !!!! NEVER COMMIT THE CREDENTIALS TO GIT !!!!!!!
################# Only demonstrated here as an example #########
################################################################
# DB username 
DB_USER = "cp84"
# DB password
DB_PWD = "neotao123"
# Database name 
DB_NAME = 'demand_acep'
#  path of timescaledb-parallel-copy
tsdb_pc_path = "/gscratch/stf/demand_acep/go/bin"

# Downsampling duration
# sample_time allows the user determine what time interval the data should be resampled at
# For 1 minute - 1T, 1 hour - 1H, 1 month - 1M, 1 Day - 1D
SAMPLE_TIME = '1T'



"""
This module contains code for the demand_acep. More documentation to come.

This is a test docstring for the whole module

"""

# %% Imports
import os
import pandas as pd
import numpy as np
import pdb
import xarray as xr
# import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from scipy.interpolate import UnivariateSpline
from itertools import groupby
from operator import itemgetter
# %%


def extract_data(dirpath, filename):
    """
        This function reads and extracts the NetCDF format data of the given meter channel using a package called xarray.

        Parameters
        ----------
        dirpath :
            `dirpath` is the directory path location of the NetCDF file to be read
        filename :
            `filename` is the NetCDF format meter channel file to be read.

        Returns
        -------
        Dataframe
            Sample-time indexed pandas dataframe containing measurement values from the given file.
        """
    # Extracts NetCDF files using xarray module
    netcdf_data = xr.open_dataset(os.path.join(dirpath, filename))
    netcdf_df = netcdf_data.to_dataframe()
    # Converts the time index to datetime format
    netcdf_df.set_index(pd.to_datetime(netcdf_df.index.values), inplace=True)

    return netcdf_df


def extract_ppty(filename, meter_name):
    """
        This function parses out the given filename as a string to determine the meter name and the measurement
        channel/type contained in the file

        Parameters
        ----------
        filename :
            `filename` is the NetCDF format meter channel file whose name contains information such as location, date,
            meter type, measurement channel/type and sampling frequency. An example filename is:
            'PokerFlatResearchRange-PokerFlat-PkFltM1AntEaDel@2018-07-02T081007Z@PT23H@PT146F.nc'
        meter_name :
            `meter_name` is a list containing the names of each of the meters at pokerflats

        Returns
        -------
        meter : string
            `meter` is the meter name of the NetCDF format file given.
        channel : string
            `channel` is the measurement type contained in the NetCDF format file given
        """
    # Example filename - 'PokerFlatResearchRange-PokerFlat-PkFltM1AntEaDel@2018-07-02T081007Z@PT23H@PT146F.nc'
    filename_split_1 = filename.split('@')
    filename_split_2 = filename_split_1[0].split('-')
    meter_channel = filename_split_2[-1]
    for name in meter_name:
        if meter_channel.startswith(name):
            n_name = len(name)
            # Example meter from example filename above - PkFltM1Ant
            meter = meter_channel[:n_name]
            # Example channel from example filename above - EaDel (Energy delivered to Phase A)
            channel = meter_channel[n_name:]

    return meter, channel


def data_resample(df, sample_time='1T'):
    """
        This function downsamples a sample-time indexed pandas dataframe containing measurement channel values based
        on the sample time supplied. It uses the mean of the values within the resolution interval. It uses the pandas
        dataframe module `df.resample`

        Parameters
        ----------
        df :
            `df` is a sample-time indexed pandas dataframe containing measurement values from the different channels of
            each meter.
        sample_time :
            `sample_time` determines the desired resolution of the downsampled data. For 1 minute - 1T, 1 hour - 1H,
            1 month - 1M, 1 Day - 1D etc. The default chosen here is 1 minute.

        Returns
        -------
        Dataframe
            Resampled-time indexed pandas dataframe containing downsampled measurement values from the given dataframe.
        """
    # Data is downsampled using the mean of the values within the interval of the sample time provided.
    # The mean is used because it provided the average/expected value of the measurement within that time range.
    df_resampled = df.resample(sample_time, closed="left", label="right").mean()

    return df_resampled


def build_interpolation(y_values, n_val):
    """
        This function takes performs the actual 1-d interpolation. If the number of consecutive missing points is less
        than 3, a linear interpolation is used, else, a cubic interpolation is used.

        Parameters
        ----------
        y_values :
            `y_values` are the values on which the function interpolation is built, that is, y_values = f(x).
        n_val :
            `n_val` is the number of consecutive missing points that needs to be filled.

        Returns
        -------
        y_interp
            Array of interpolated values equal in length to the missing supplied length (n_val) of missing data points..
        """
    # removes the NaN values so as not to skew the performance of the scipy interp1d function.
    y_values = y_values.dropna()
    x = np.linspace(1, len(y_values), num=len(y_values))
    # x = np.reshape(x, y_values.shape)
    y_values = np.asarray(y_values).squeeze()
    # pdb.set_trace()
    # if-else uses a linear interpolation when the number of consecutive missing data points is less than 3.
    #  The number of points to be interpolated has to be greater than 3 points to use the spline/Cubic interpolation
    if len(y_values) <= 3:
        f = interp1d(x, y_values, kind='linear')
    else:
        f = interp1d(x, y_values, kind='cubic')

    x_interp = np.linspace(1, len(y_values), num=n_val)
    y_interp = f(x_interp)

    return y_interp


def compute_interpolation(df):
    """
        This function imputes missing measurement data (Nan) in a series using 1-d interpolation.

        Parameters
        ----------
        df :
            `df` is a series containing missing measurements values.

        Returns
        -------
        Series
            Filled pandas series with no missing values.
        """
    # creates a deep copy of the Series received
    test_df = df.copy()
    # gets the index location in integers where the NaNs are located
    get_nan_idx = np.where(test_df.isna())[0]
    idx_grp_nan = []
    # creates a list of consecutive index locations to determine the range of interpolation
    for k, g in groupby(enumerate(get_nan_idx), lambda ix: ix[0] - ix[1]):
        idx_grp_nan.append(list(map(itemgetter(1), g)))
    # performs interpolation for each consecutive NaN index location
    for idx, val in enumerate(idx_grp_nan):
        n_grp = len(val)
        # for each range of consecutive NaN locations, use data points of length equal to the number that NaNs in that
        # range before and after the NaN data points
        prev_idx = val[0] - n_grp
        next_idx = val[-1] + n_grp
        # This if-else clause handles edge cases
        # If - When the number of consecutive NaN points is larger than the number of available data points before it in
        # the dataframe.
        # Elif - When the number of consecutive NaN points is larger than the number of available data points after it
        # in the dataframe.
        if prev_idx < 0:
            prev_vals = test_df.iloc[0:val[0]]
            next_vals = test_df.iloc[val[0] + 1: next_idx + 1]
        elif next_idx > len(test_df):
            prev_vals = test_df.iloc[prev_idx:val[0]]
            next_vals = test_df.iloc[val[0] + 1: len(test_df)]
        else:
            prev_vals = test_df.iloc[prev_idx:val[0]]
            next_vals = test_df.iloc[val[0] + 1: next_idx + 1]
        y_values = prev_vals.append(next_vals)
        y_interp = build_interpolation(y_values, n_grp)
        test_df.iloc[val] = y_interp

    return test_df


def data_impute(impute_df):
    """
        This function imputes missing measurement in a dataframe using a 1-d interpolation. If the number of consecutive
        missing points is less than 3, a linear interpolation is used, else, a cubic interpolation is used.

        Parameters
        ----------
        impute_df :
            `impute_df` can either be a dataframe of a dictionary of dataframes containing missing measurements values.

        Returns
        -------
        Dataframe
            Filled pandas dataframe with no missing values.
        """
    # if-else checks if the input is a dataframe or list of dataframes.
    if isinstance(impute_df, dict):
        for meter in impute_df:
            if impute_df[meter].isnull().values.any():
                impute_df[meter] = impute_df[meter].apply(compute_interpolation)
    else:
        if impute_df.isnull().values.any():
            impute_df = impute_df.apply(compute_interpolation)

    return impute_df


def long_missing_data_prep(dirpath, filename):
    """
        This function prepares a dataset in a `csv` format with missing days, months or years for interpolation
        using the `data_impute` function. It fills in the missing time as a 'DateTimeIndex' and assigns a value
        of NaN to the missing data points.

        Parameters
        ----------
        dirpath :
            `dirpath` is the directory path location of the csv file containing the missing data points in already
            down-sampled to a 1-Minute interval.
        filename :
            `filename` is the csv file containing the missing data points to be read.

        Returns
        -------
        Dataframe
            pandas dataframe with 'DateTimeIndex' and value of NaN assigned to the missing data points.
        """
    # Reads dataset into a Pandas Dataframe
    meter_df = pd.read_csv(os.path.join(dirpath, filename))
    # Converts the time column to a DateTimeIndex format
    meter_df.set_index(pd.to_datetime(meter_df['time']), inplace=True)
    st_time = meter_df['time'].iloc[0]
    sp_time = meter_df['time'].iloc[-1]
    # Generates a date time range at an interval of 1 minute from the start and end date of the dataset.
    date_idx = pd.date_range(start=st_time, end=sp_time, freq='1T')
    meter_df.drop(columns=['time'], inplace=True)
    # Reindex the dataframe to include the DateTimeIndex for the missing days and NaN values for missing points
    meter_mod_df = meter_df.reindex(date_idx)

    return meter_mod_df






""" This file contains functions that create the schema in the database """

from sqlalchemy import *
import pandas as pd
import os

def create_schema_from_source_files(sql_engine, config):
    """ 
    This function reads the source file and extracts the table names and
    corresponding column names.
    
    Parameters
    ----------
    sql_engine : SQLAlchemy engine
        'sql_engine` should support database operation.
    channel_metadata_file : string
        `channel_metadata_file` represents the absolute path of the channel 
        metadata file **including** the filename.
    years_file : string
        `years_file` represents the absolute path of the file containing the 
        years for which the tables should be created for. This file should be a 
        text file with one year per row and the column header "years". 

    Returns
    -------
    List of table names in the database.
    """
    
    # meter_channel_dict, data_years = read_source_files(channel_metadata_file, years_file)
    
    # Metadata for the SQLAlchemy tables database - will add tables eventually 
    # This should be connected to the correct database. 
    metadata = MetaData(sql_engine)
    metadata.reflect(sql_engine)
    
    delete_ok = input("This will drop all the tables and delete all the data, OK? Enter y or n")
    # Clean the database, 
    # ***** this will delete all tables and data in it *****
    
    if delete_ok.lower() == 'y':
        metadata.drop_all(sql_engine)
        # Clear the metadata
        metadata.clear()
    
        for meter in config.METER_CHANNEL_DICT:
            # Get the channels for this meter
            meter_channels = config.METER_CHANNEL_DICT[meter]
            # Create the following lists for the database insertion 
            # Types of columns - defaulting to float, except for time
            columns_types = [TIMESTAMP(timezone=True)] + [Float] * (len(meter_channels) - 1)
            # Primary key flags - defaulting to false, except for time
            # **** removing the pk_constraint, for speedy inserts ****
            primary_key_flags = [True] + [False] * (len(meter_channels) - 1)
            # Nullable flags - defaulting to false, except for time
            # **** removing the non_null constraint, for speedy inserts ****
            nullable_flags = [True] + [False] * (len(meter_channels) - 1)
            # Iterate over the years list 
            for year in config.DATA_YEARS: 
                # Create table name from meter name and year 
                table_name = meter + "_" + str(year)
                
                # Dynamically create table names and corresponding column names
                table_name_i = Table(table_name, metadata,
                 *(Column(column_name, column_type,
                          primary_key=primary_key_flag,
                          nullable=nullable_flag)
                   for column_name,
                       column_type,
                       primary_key_flag,
                       nullable_flag in zip(meter_channels,
                                            columns_types,
                                            primary_key_flags,
                                            nullable_flags)))
                print(table_name_i)
        
        # Create the tables in the metadata - this defaults to checking that 
        # the table names first, so will only create tables if they do not exist
        # making this function idempotent
        metadata.create_all()
        
        # Create timescaledb hypertables from the tables created 
        
        ##########
        # There is currently no SQLAlchemy way to create hypertables, so we execute 
        # an sql statement. This would have been simlar in psycopg2. 
        # Also, this creates hypertables for all tables in the metadata. If this is 
        # desired, then specify the tables names using some other mechanism. 
        #########
        
        # First get the names of the tables 
        table_names = metadata.tables.keys()
    
        # Connect to the database engine to execute the query
        with sql_engine.connect() as con:
            
            # Iterate over the table names to create the query statement
            for table_name in table_names:
                # Generate the query statement for the current table name
                sql_statement = ("""SELECT create_hypertable('public.""" + '"' + 
                table_name + '"' + """', 'time' , if_not_exists => TRUE);""")
                # Execute the query 
                con.execute(sql_statement)
    
    return  metadata.tables.keys()



"""
Functions in this file insert the data into the timescaledb database using the 
go utility provided by timescaledb called timescaledb-parallel-copy
"""

import os
import pandas as pd
import pdb
import datetime
import io
import glob
import multiprocessing
import subprocess

from demand_acep import extract_data
from demand_acep import extract_ppty

def parallel_copy_data_for_date(config, data_date):
    """ 
    This function parallel copies the data for the data_date specified to the 
    appropriate table of TSDB.
    
    Parameters
    ----------
    config : 
        `config` contains the configuration for paths etc. needed by files. As
        an argument we can change the config files for different production 
        vs test configs 
    data_date : string
        `data_date` string will be used to extract the year and the path to the
        data. It should be in the 'mm/dd/yyyy' format. 

    Returns
    -------
    int
        Description of anonymous integer return value.
    """    
    print(data_date)
    # Look for the pickle and file and use pandas_to_sql to insert the data 
    # into the database. 
    data_year = data_date[-4:]
    # print(data_year)
    # print(config.DATA_ROOT)
    # find this path, extract data from those files, create a dataframe, and 
    # store it as pickle in the folder. 
    data_month = data_date[0:2]
    data_day = data_date[3:5]
    data_path = os.path.join(config.DATA_ROOT, data_year, data_month, data_day)
    # print(data_path)
    # print(config.METER_CHANNEL_DICT)
    # print(config.DATA_YEARS)
    
    meter_count = len(config.METER_CHANNEL_DICT)
    # Create a list of lists of size of number of meters
    meter_collection = {}
    meter_pickle_names = {}
    # Path of the tsdb parallel copy go executable
    timescaledb_parallel_copy_path = os.path.join(config.tsdb_pc_path, "timescaledb-parallel-copy")
    db_address = config.DB_ADDRESS
    db_name = config.DB_NAME
    db_username = config.DB_USER
    db_pwd = config.DB_PWD
    # Get the CPU count for parallelizing the process
    num_workers = multiprocessing.cpu_count()
    # Generate the connection string for the database
    connection_string = 'host={0} user={1} password={2} sslmode=disable'.format(db_address, db_username, db_pwd)
    
    
    for fname in glob.glob(os.path.join(data_path, '*.csv')):
        # print(os.path.basename(fname))
        file_name = os.path.basename(fname)
        table_name = file_name.split('@')[0] + '_' + data_year
        parallel_copy_cmd = [timescaledb_parallel_copy_path, '--db-name', 
                            db_name, '--table', table_name, '--file', fname, 
                            '--workers', str(num_workers), '--connection', 
                            connection_string, '--skip-header', 'true']
        # print(parallel_copy_cmd)
        pcopy = subprocess.run(parallel_copy_cmd, encoding='utf-8', stdout=subprocess.PIPE)
        for line in pcopy.stdout.split('\n'):
            print(line)

    return 

def parallel_copy_data_for_dates(config, start, end):
    """ 
    This function extracts the data to disk for the data_date_range specified.
    
    Parameters
    ----------
    config : 
        `config` contains the configuration for paths etc. needed by files. As
        an argument we can change the config files for different production 
        vs test configs 
    start : str or datetime-like
        Left bound for generating dates.
    end : str or datetime-like
        Right bound for generating dates.
 
    Returns
    -------
    int
        Description of anonymous integer return value.
    """
    
    # Get the list of dates within the range and convert to the format we want
    dates_in_range = pd.date_range(start, end).strftime("%m/%d/%Y")
    print(dates_in_range)
    # Perform the insert for the said days
    for date in dates_in_range:
        parallel_copy_data_for_date(config, date)

    return 
    
    